/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.checker;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import fr.supelec.adlv.ActivateComponentStep;
import fr.supelec.adlv.ActiveExpression;
import fr.supelec.adlv.AlwaysClause;
import fr.supelec.adlv.Application;
import fr.supelec.adlv.BinaryExpression;
import fr.supelec.adlv.BinaryOperator;
import fr.supelec.adlv.Component;
import fr.supelec.adlv.ConfigurationStep;
import fr.supelec.adlv.Connection;
import fr.supelec.adlv.ConnectionConfigurationStep;
import fr.supelec.adlv.ConsumerPort;
import fr.supelec.adlv.ControlComponentType;
import fr.supelec.adlv.DeactivateComponentStep;
import fr.supelec.adlv.Expression;
import fr.supelec.adlv.InternalComponentType;
import fr.supelec.adlv.LiteralExpression;
import fr.supelec.adlv.NeverClause;
import fr.supelec.adlv.Port;
import fr.supelec.adlv.PortExpression;
import fr.supelec.adlv.ProducerPort;
import fr.supelec.adlv.SafetyClause;
import fr.supelec.adlv.UnaryExpression;
import fr.supelec.adlv.UnaryOperator;
import fr.supelec.adlv.checker.logic.ActivePredicate;
import fr.supelec.adlv.checker.logic.Always;
import fr.supelec.adlv.checker.logic.And;
import fr.supelec.adlv.checker.logic.BackTo;
import fr.supelec.adlv.checker.logic.BinaryExpr;
import fr.supelec.adlv.checker.logic.ComparisonOperator;
import fr.supelec.adlv.checker.logic.ComparisonPredicate;
import fr.supelec.adlv.checker.logic.ConnectedPredicate;
import fr.supelec.adlv.checker.logic.Constant;
import fr.supelec.adlv.checker.logic.Expr;
import fr.supelec.adlv.checker.logic.Implies;
import fr.supelec.adlv.checker.logic.Interval;
import fr.supelec.adlv.checker.logic.Not;
import fr.supelec.adlv.checker.logic.Once;
import fr.supelec.adlv.checker.logic.Or;
import fr.supelec.adlv.checker.logic.Previous;
import fr.supelec.adlv.checker.logic.ProtoInterval;
import fr.supelec.adlv.checker.logic.SignalPredicate;
import fr.supelec.adlv.checker.logic.Since;
import fr.supelec.adlv.checker.logic.UnaryExpr;
import fr.supelec.adlv.checker.logic.Expr.ExprBundle;
import fr.supelec.adlv.checker.logic.Expr.Relation;
import fr.supelec.adlv.parser.EmfAdlvHelper;


/**
 * This class represents fulfillment and un-fulfillment signals for a given
 * "active" or "inactive" condition.
 *
 */
class Conditions {
	public Set<PortInstance> enablingPorts = new HashSet<PortInstance>();
	public Set<PortInstance> disablingPorts = new HashSet<PortInstance>();
	public boolean initiallyEnabled = false;
	public boolean isStatic = false;

	@Override
	public String toString() {
		StringBuffer buf = new StringBuffer("\tInitially enabled: " + initiallyEnabled + "\n");
		buf.append("\tStatic: ").append(isStatic).append('\n');
		buf.append("\tEnabling:  ").append(enablingPorts).append('\n');
		buf.append("\tDisabling: ").append(disablingPorts).append('\n');
		return buf.toString();
	}

	public void invert() {
		initiallyEnabled = !initiallyEnabled;
		Set<PortInstance> temp = enablingPorts;
		enablingPorts = disablingPorts;
		disablingPorts = temp;
	}
}


class ElementConfiguration {

}


class ConnectionConfiguration extends ElementConfiguration {
	public final PortInstance from;
	public final PortInstance to;

	public ConnectionConfiguration(PortInstance from, PortInstance to) {
		this.from = from;
		this.to = to;
	}

	@Override
	public String toString() {
		return "(dis)connection of \"" + from + " -> " + to + "\"";
	}

	@Override
	public boolean equals(Object o) {
		if(o instanceof ConnectionConfiguration) {
			ConnectionConfiguration cco = (ConnectionConfiguration)o;
			return from == cco.from && to == cco.to;
		}
		return false;
	}

	@Override
	public int hashCode() {
		return from.hashCode() ^ to.hashCode();
	}
}


class ComponentConfiguration extends ElementConfiguration {
	public final Component component;

	public ComponentConfiguration(Component component) {
		this.component = component;
	}

	@Override
	public String toString() {
		return "(de)activation of component \"" + component.getName() + "\"";
	}

	@Override
	public boolean equals(Object o) {
		if(o instanceof ComponentConfiguration) {
			ComponentConfiguration cco = (ComponentConfiguration)o;
			return component == cco.component;
		}
		return false;
	}

	@Override
	public int hashCode() {
		return component.hashCode();
	}
}


class HasseNode {
	public final PortInstance port;
	public final List<HasseRelation> relations;
	public final Expr expr;

	public HasseNode(PortInstance port, Expr expr) {
		this.port = port;
		this.expr = expr;
		relations = new ArrayList<HasseRelation>();
	}

	public void addRelation(HasseRelation rel) {
		if(rel.type == HasseRelation.Type.EQUIV) {
			relations.add(rel);
			return;
		}


		// type SUCC ou PREC
		for(HasseRelation aRelation : relations) {
			HasseRelation.Type comp = HasseRelation.fromImplicationRelation(
					aRelation.toNode.expr.compare(rel.toNode.expr));
			////System.out.println("\t\t\tComparing: \"" + aRelation.toNode.expr + "\" " + comp + " \" " + rel.toNode.expr + "\"");
			if(comp == aRelation.type) {
				aRelation.toNode.addRelation(rel);
				return;
			}

			if(rel.type == aRelation.type &&
					((comp == HasseRelation.Type.SUCC && aRelation.type == HasseRelation.Type.PREC) ||
					(comp == HasseRelation.Type.PREC && aRelation.type == HasseRelation.Type.SUCC))) {
				relations.remove(aRelation);
				relations.add(rel);
				rel.toNode.addRelation(aRelation);
				return;
			}
		}

		// else add the relation
		relations.add(rel);
	}

	public void addRelation(HasseRelation.Type type, HasseNode toNode) {
		addRelation(new HasseRelation(type, toNode));
	}

	public void addRelation(HasseRelation.Type type, PortInstance port, Expr expr) {
		addRelation(type, new HasseNode(port, expr));
	}

	public void listEquivPrecSucc(Set<Signal> equiv, Set<Signal> prec, Set<Signal> succ) {
		for(HasseRelation rel : relations) {
			Set<Signal> set;
			switch(rel.type) {
			case EQUIV: set = equiv; break;
			case PREC: set = prec; break;
			case SUCC: set = succ; break;
			default: continue;
			}
			set.add(new Signal(rel.toNode.port, true));
		}
	}

	@Override
	public String toString() {
		StringBuffer res = new StringBuffer("[");
		if(port != null) res.append(port).append(": ");
		res.append(Util.collectionToString(relations));
		res.append(']');
		return res.toString();
	}
}

class HasseRelation {
	public final Type type;
	public final HasseNode toNode;

	public HasseRelation(Type type, HasseNode toNode) {
		this.type = type;
		this.toNode = toNode;
	}

	@Override
	public String toString() {
		return type.toString() + toNode.toString();
	}

	public static HasseRelation.Type fromImplicationRelation(Relation r) {
		switch(r) {
		case EQUIV: return Type.EQUIV;
		case IMPLIES: return Type.SUCC;
		case IS_IMPLIED: return Type.PREC;
		default: return null;
		}
	}

	public enum Type {
		//PREC("≤"), SUCC("≥"), EQUIV("↔");
		PREC("<"), SUCC(">"), EQUIV("<->");

		private final String symbol;

		private Type(String symbol) {
			this.symbol = symbol;
		}

		@Override
		public String toString() {
			return symbol;
		}
	}
}


/**
 * This class represents fulfillment and un-fulfillment signals for a given
 * boolean condition.
 *
 */
class SignalsEmittedWhen {
	public final HasseNode whenTrue = new HasseNode(null, null);
	public final HasseNode whenFalse = new HasseNode(null, null);

	@Override
	public String toString() {
		StringBuffer buf = new StringBuffer("SEW\tTrue:  ").append(whenTrue).append("\n");
		buf.append("   \tFalse: ").append(whenFalse).append("\n");
		return buf.toString();
	}
}



/**
 * This class is responsible for generating checking code (in a given language)
 * for safety conditions about an ADLV description.
 *
 */
public class Checker {
	private final Application app;
	private final Component controlComponent;
	private final LogFacility log;

	private final Map<ConnectionConfiguration, Conditions> enableDisableConnections =
		new HashMap<ConnectionConfiguration, Conditions>();

	private final Map<Component, Conditions> activateDeactivateComponents =
		new HashMap<Component, Conditions>();

	private final Map<ElementConfiguration, Conditions> configConditions =
		new HashMap<ElementConfiguration, Conditions>();


	public Checker(Application app, LogFacility log) {
		this.app = app;
		this.controlComponent = findControlComponent();
		this.log = log;

		// scan enable/disable, connect/disconnect steps

		// 1) Static configuration
		for(ConfigurationStep ostep : app.getStaticConfiguration()) {
			addStep(ostep, true, true, null);
		}


		// 2) initial dynamic configuration
		for(ConfigurationStep ostep : app.getInitialSteps()) {
			addStep(ostep, false, true, null);
		}


		// 3) transitions for dynamic configuration
		for(Port oPort : app.getPorts()) {
			if(oPort instanceof ConsumerPort) {
				ConsumerPort cp = (ConsumerPort)oPort;
				for(ConfigurationStep ostep : cp.getSteps()) {
					addStep(ostep, false, false, portFromControllerPOV(cp));
				}
			}
		}

		// 4) post-process: calculate deactivating conditions for connections
		for(Map.Entry<ElementConfiguration, Conditions> e : configConditions.entrySet()) {
			ElementConfiguration eltConfig = e.getKey();
			// for each connection configuration entry...
			if(eltConfig instanceof ConnectionConfiguration) {
				ConnectionConfiguration config = (ConnectionConfiguration)eltConfig;
				Conditions cond = e.getValue();
				// add disabling conditions for all configuration entries with
				// the same "to" port
				for(Map.Entry<ElementConfiguration, Conditions> te : configConditions.entrySet()) {
					if(te.getKey() instanceof ConnectionConfiguration) {
						ConnectionConfiguration tConfig = (ConnectionConfiguration)te.getKey();
						Conditions tCond = te.getValue();

						// all target entries with same "to" port (but not the
						// same "from" port as it would be oneself)
						if(tConfig.to == config.to && tConfig.from != config.from) {
							// add all the current entry's enabling conditions
							// as disabling conditions for the target entry
							tCond.disablingPorts.addAll(cond.enablingPorts);
						}
					}
				}
			}

		}

		// 5) print out the results AND fill the lists :
		//    enableDisableConnections and activateDeactivateComponents

		log.debug("******** Configuration conditions ********");
		for(Map.Entry<ElementConfiguration, Conditions> e : configConditions.entrySet()) {
			ElementConfiguration config = e.getKey();
			Conditions cond = e.getValue();
			log.debug(config);
			log.debug(cond);

			if(config instanceof ConnectionConfiguration)
				enableDisableConnections.put((ConnectionConfiguration)config, cond);
			if(config instanceof ComponentConfiguration)
				activateDeactivateComponents.put(((ComponentConfiguration)config).component, cond);
		}
		log.debug("******************************************");
	}


	/**
	 * This function is to be called by the constructor, when scanning for
	 * enable/disable and connect/disconnect steps.
	 * @param ostep
	 * @param isStatic
	 * @param isInitial
	 * @param ctrlPort
	 */
	private void addStep(ConfigurationStep ostep, boolean isStatic, boolean isInitial, PortInstance ctrlPort) {
		// 1) what is this step?
		//      - enable or disable?
		//      - component (de)activation or (dis)connection?
		ElementConfiguration config = null;
		boolean enable = false;

		if(ostep instanceof ActivateComponentStep) {
			config = new ComponentConfiguration(((ActivateComponentStep)ostep).getComponent());
			enable = true;
		}
		if(ostep instanceof DeactivateComponentStep) {
			config = new ComponentConfiguration(((DeactivateComponentStep)ostep).getComponent());
			enable = false;

		}
		if(ostep instanceof ConnectionConfigurationStep) {
			ConnectionConfigurationStep ccs = (ConnectionConfigurationStep)ostep;
			Connection connection = ccs.getConnection();
			config = new ConnectionConfiguration(
					PortInstance.make(connection.getFromPort(), connection.getFromComponent()),
					PortInstance.make(connection.getToPort(), connection.getToComponent()));
			enable = true;

			// at first, all connection configuration steps are treated as
			// enabling operations. A later pass will calculate the
			// corresponding disabling conditions.
		}

		if(config == null)
			log.err("Unrecognized configuration step: " + ostep);

		// 2) already another condition for the same configuration?
		Conditions cond = configConditions.get(config);

		if(cond==null) { // no -> new condition
			cond = new Conditions();
			cond.isStatic = isStatic;
			if(isInitial) {
				cond.initiallyEnabled = enable;
			} else {
				if(enable) cond.enablingPorts.add(ctrlPort);
				else cond.disablingPorts.add(ctrlPort);
			}
		} else { // yes -> update condition
			if(isInitial || isStatic) log.err("Several initial configurations for " + config);
			if(cond.isStatic) log.err("Changes to STATIC configuration for " + config);
			if(enable) cond.enablingPorts.add(ctrlPort);
			else cond.disablingPorts.add(ctrlPort);
		}

		// put back the new or modified conditions
		configConditions.put(config, cond);
	}


	private Component findControlComponent() {
		for(Component comp : app.getComponents()) {
			if(comp.getType() instanceof ControlComponentType) {
				return comp;
			}
		}
		log.err("Could not find control component.");
		return null;
	}


	/**
	 * Transforms an Expression (EMF model) into an Expr (model used
	 * internally by the checker generator).
	 * @param te
	 * @return
	 */
	private Expr expressionToExpr(Expression te, Component currentComponent) {
		if(te == null) {
			log.err("expressionToExpr() called on a null expression. This should not happen.");
			return null;
		}

		// Form: active(component)
		if(te instanceof ActiveExpression) {
			ActiveExpression ap = (ActiveExpression)te;
			return new ActivePredicate(ap.isActive(), ap.getComponent());
		}

		// Form: expr1 BINOP expr2 where BINOP can be:
		//   - a comparison operator: <, <=, >, etc.
		//   - a logical operator: and, or, since, etc.
		//   - the "connected" operator: <<
		if(te instanceof BinaryExpression) {
			BinaryExpression be = (BinaryExpression)te;
			Expression left = be.getLeftExpression();
			Expression right = be.getRightExpression();
			BinaryOperator binOp = be.getOperator();

			switch(BinaryOperatorClassifier.classify(binOp)) {
			case COMPARISON:
				Expression[] lre = {left, right};

				PortInstance port = null;
				String cst = null;

				for(Expression se : lre) {
					if(se instanceof LiteralExpression) {
						cst = ((LiteralExpression)se).getValue();
					}
					if(se instanceof PortExpression) {
						PortExpression pe = (PortExpression)se;
						Port portType = pe.getPort();
						Component comp = pe.getComponent();

						// If inside a "when" clause, portType.getComponentType()
						// is of the right type, but pe.getComponent() is null.
						// This is because the "when" clause is defined at the
						// level of component types, not at the level of
						// component instances. So we need to set the right
						// component in this case.
						// The same applies below in this function.
						port = PortInstance.makeWithCurrent(
								portType, comp, currentComponent);
					}
				}

				if(port == null  || cst == null) {
					log.err("Could not create an Expr for comparison: "
							+ EmfAdlvHelper.getStringForExpression(te),
							"We expect an expr to be a literal, and the other to be a port.");
					return null;
				}

				ComparisonOperator op;
				switch(binOp) {
				case EQUAL: op = ComparisonOperator.EQUAL; break;
				case NOT_EQUAL: op = ComparisonOperator.NEQ; break;
				case GREATER: op = ComparisonOperator.GREATER; break;
				case GREATER_EQUAL: op = ComparisonOperator.GEQ; break;
				case LESS: op = ComparisonOperator.LESS; break;
				case LESS_EQUAL: op = ComparisonOperator.LEQ; break;
				default: log.err("In " + EmfAdlvHelper.getStringForExpression(be) +
						": unknonwn CompareOperator: " + binOp); return null;
				}

				return new ComparisonPredicate(port, op, cst);

			case LOGICAL:
				Expr eLeft = expressionToExpr(left, currentComponent);
				Expr eRight = expressionToExpr(right, currentComponent);
				switch(binOp) {
				case AND: return new And(eLeft, eRight);
				case OR: return new Or(eLeft, eRight);
				case IMPLIES: return new Implies(eLeft, eRight);
				case BACK_TO: return new BackTo(eLeft, eRight);
				case SINCE: return new Since(eLeft, eRight);
				default: log.err("In " + EmfAdlvHelper.getStringForExpression(be) +
						": unknown logical binary operator: " + binOp);
				}

			case CONNECTED:
				if(!(left instanceof PortExpression) || !(right instanceof PortExpression)) {
					log.err("Connected predicate between two expressions that are not port expressions: " +
							EmfAdlvHelper.getStringForExpression(be));
					return null;
				}

				PortExpression peRight = (PortExpression)right;
				PortExpression peLeft = (PortExpression)left;
				return new ConnectedPredicate(
						PortInstance.makeWithCurrent(peRight.getPort(), peRight.getComponent(), currentComponent),
						PortInstance.makeWithCurrent(peLeft.getPort(), peLeft.getComponent(), currentComponent),
						true);

			default:
				log.err("In " + EmfAdlvHelper.getStringForExpression(be) +
						": unknown binary operator: " + binOp);
			return null;
			}
		}


		// Form: UNARY_OP expr
		if(te instanceof UnaryExpression) {
			UnaryExpression ue = (UnaryExpression)te;
			UnaryOperator unOp = ue.getOperator();
			Expression e = ue.getExpression();
			Expr expr = expressionToExpr(e, currentComponent);

			switch(unOp) {
			case NONE: return expr; // do nothing (parenthesized expression)
			case NOT: return new Not(expr);
			case ALWAYS: return new Always(expr);
			case ONCE: return new Once(expr);
			case PREVIOUS: return new Previous(expr);
			default: log.err("In " + EmfAdlvHelper.getStringForExpression(ue) +
					": unknown unary operator: " + unOp);
			}
		}

		// Form: <port name>
		if(te instanceof PortExpression) {
			PortExpression pe = (PortExpression)te;
			return new SignalPredicate(true,
					PortInstance.makeWithCurrent(pe.getPort(), pe.getComponent(), currentComponent));
		}


		// Form: <literal>
		if(te instanceof LiteralExpression) {
			log.err("Literal expression \"" + EmfAdlvHelper.getStringForExpression(te) +
					"\" should not be used as a first-class expression.");
			return null;
		}


		log.err("Could not create an Expr for: " +
				EmfAdlvHelper.getStringForExpression(te) +
				"of class: " + te.getClass().getCanonicalName());

		return null; // fall-through case
	}


	/**
	 * Finds the signals that cause the activation or deactivation of a given
	 * component.
	 *
	 * @param compId the identifier of the component
	 * @return the activating and deactivating conditions
	 */
/*	private Conditions findActivatingDeactivatingPorts(String compId) {
		Component comp = helper.getApplicationComponent(compId);
		if(comp == null) return null;
		return findActivatingDeactivatingPorts(comp);
	}*/


	/**
	 * Returns the signals that cause the activation or deactivation of a given
	 * component.
	 *
	 * @param comp the component
	 * @return the activating and deactivating conditions
	 */
	private Conditions findActivatingDeactivatingPorts(Component comp) {
		return activateDeactivateComponents.get(comp);
	}


	/**
	 * Returns the signals that cause the connection or deconnection of two
	 * ports.
	 *
	 * @param portFrom source port
	 * @param portTo destination port
	 * @return the connection and deconnection conditions
	 */
	private Conditions findConnectionDisconnectionPorts(PortInstance portFrom, PortInstance portTo) {
		return enableDisableConnections.get(new ConnectionConfiguration(portFrom, portTo));
	}


	/**
	 * Finds which are the signals emitted when a given condition occurs
	 * (i.e. appears in a "where" clause).
	 *
	 * @param expr an expression representing the condition
	 * @return
	 */
	private SignalsEmittedWhen findSignalsEmittedWhen(Expr expr) {
		SignalsEmittedWhen sew = new SignalsEmittedWhen();

		log.debug("findSignalsEmittedWhen " + expr);

		Expr negExpr = new Not(expr);

		log.debug("\tPositive: " + expr);
		log.debug("\tNegative: " + negExpr);

		// iterate over INTERNAL components
		for(Component comp : app.getComponents()) {
			if(! (comp.getType() instanceof InternalComponentType)) continue;

			log.debug("\t->Internal component: " + comp.getName() + "  type=" + comp.getType().getName());

			// iterate over the component's producer ports
			for(Port oPort : comp.getType().getPorts()) {
				if(oPort instanceof ProducerPort) {
					ProducerPort port = (ProducerPort)oPort;
					Expression emitExpr = port.getEventEmitExpression();

					// if no when-condition then skip
					if(emitExpr == null) continue;

					log.debug("\texamining " + comp.getName() + "::" + port.getName() +
							" (when " + EmfAdlvHelper.getStringForExpression(emitExpr) + ")");
					Expr eEmitExpr = expressionToExpr(emitExpr, comp);
					if(eEmitExpr == null) continue;

					// process positive and negative versions
					for(int i=0; i<2; i++) {
						HasseRelation.Type type = HasseRelation.fromImplicationRelation(
								(i==0 ? expr : negExpr).compare(eEmitExpr));


						if(type != null) {
							log.debug("\t\tFound relation for " +
									(i==0 ? "positive" : "negative") +
									(" version: ") + type);
							(i==0 ? sew.whenTrue : sew.whenFalse).addRelation(type,
									portFromControllerPOV(port),
									eEmitExpr);
							////log.debug("\t\t" + sew);
						}
					}
				}
			}
		}
		return sew;
	}


	/**
	 * Returns the controller port corresponding to a given port. This port
	 * must therefore be connected to the controller.
	 *
	 * This methods returns a port, as seen from the controller's point of
	 * view, hence the method name.
	 * @param port
	 * @return
	 */
	private PortInstance portFromControllerPOV(Port port) {
		for(Connection cnx : port.getConnectionTo()) {
			// port is a consumer port here

			if(cnx.getFromComponent() == controlComponent)
				return PortInstance.make(cnx.getFromPort(), controlComponent);
		}

		for(Connection cnx : port.getConnectionFrom()) {
			// port is a producer port here

			if(cnx.getToComponent() == controlComponent)
				return PortInstance.make(cnx.getToPort(), controlComponent);
		}

		log.err("Port not directly connected to controller: " + port);
		return null;
	}



	/**
	 * Generates checking code for a "never" clause.
	 */
	private void generateCheckingCodeForNeverClause(CheckingCodeGenerator codeGen, Expr clauseExpr) {
		final String logPrefix = "[gCCFNeverClause] ";
		log.debug(logPrefix + "Processing 'never' clause: " + clauseExpr);

		log.debug("% For clause: " + clauseExpr);

		Expr protoPattern = patternForClause(clauseExpr, true);
		log.debug("% Proto-pattern: " + protoPattern);

		ExprBundle bundle = protoPattern.getBundle();

		Expr[] patterns;
		String[] descriptions;
		String[] longdesc;

		if(bundle.isBundled) {
			log.info("Could not find exact matches for 'never' clause " + clauseExpr);
			log.info("Generating approximate observers: " +
					(bundle.inner != null ? "inner (too weak)   " : "") +
					(bundle.outer != null ? "outer (too strong)   " : ""));
			longdesc = new String[] {
					"Too weak pattern: the check detects only legitimous errors, but if it passses, it proves nothing.",
					"Too strong pattern: if the check passes, the program is OK, but the check may fail in false positive cases"
			};

			if(bundle.outer != null)
				log.info( longdesc[1] );
			if(bundle.inner != null)
				log.info( longdesc[0] );
			patterns = new Expr[] { bundle.inner, bundle.outer };
			descriptions = new String[] { "too weak (inner)", "too strong (outer)" };
		} else {
			patterns = new Expr[] { protoPattern };
			descriptions = new String[] { "exact" };
			longdesc = new String[] { "Exact pattern, equivalent to the specified property." };
		}

		for(int i = 0; i<patterns.length; i++) {
			Expr pattern = patterns[i];
			if(pattern == null) continue;

			String descr = descriptions[i];

			log.debug("% Raw " + descr + " pattern: " + pattern);

			// try and simplify patterns
			Expr simplePattern = pattern.simplify();
			log.debug("% Simplified " + descr + " pattern: " + simplePattern);

			if(Expr.FALSE.equals(simplePattern)) {
				// optimization: no need to generate checking code is the Never
				// clause is of the form " Never { false } "
				log.info("Clause \"" + clauseExpr + "\" yields (" + descr + ") \"" + pattern +
						"\" which is equivalent to \"" + simplePattern +
						"\", which is useless. Not generating it.");
			} else {
				if(Expr.TRUE.equals(simplePattern)) {
					log.info("Clause \"" + clauseExpr + "\" yields (" + descr + ") \"" + pattern +
							"\" which is equivalent to \"" + simplePattern +
							"\", which will ALWAYS FAIL...");
					log.info("Generating checking code anyway.");
				}

				// general case
				log.debug("% Now generating checking code");
				codeGen.generateNeverClause(pattern, new String[] {
						"For never clause:",
						"\tnever { " + clauseExpr + " }",
						longdesc[i],
						"\t" + descriptions[i] + " pattern: " + pattern
				});
			}
		}
	}


	/**
	 * Returns a pattern (intermediate form, in which only appear "interval"
	 * predicates and logical operators) for a given Expr.
	 *
	 * @param clause
	 */
	private Expr patternForClause(Expr clause, boolean abortIfNoSolution) {
		// the expression is of the form "active(component)" or "inactive(component)"
		if(clause instanceof ActivePredicate) {
			ActivePredicate ap = (ActivePredicate)clause;
			Component comp = ap.getComponent();
			Conditions edc = findActivatingDeactivatingPorts(comp);

			if(!ap.isActive()) edc.invert();

			log.debug("== For active predicate: " + ap);
			log.debug(edc);

			if(edc.enablingPorts.size() == 0 && edc.disablingPorts.size() == 0) {
				return new Constant(edc.initiallyEnabled);
			}

			return new Interval(
					Signal.positiveSignalsFromPorts(edc.enablingPorts),
					Interval.SignalCombination.OR,
					Signal.positiveSignalsFromPorts(edc.disablingPorts),
					Interval.SignalCombination.OR,
					edc.initiallyEnabled);
		}

		// the expression is a "connected predicate" (port_a << port_b)
		if(clause instanceof ConnectedPredicate) {
			ConnectedPredicate cp = (ConnectedPredicate)clause;
			Conditions edc = findConnectionDisconnectionPorts(cp.getFromPort(), cp.getToPort());

			if(!cp.isConnected()) edc.invert();

			log.debug("== For connected predicate: " + cp);
			log.debug(edc);

			if(edc.enablingPorts.size() == 0 && edc.disablingPorts.size() == 0) {
				return new Constant(edc.initiallyEnabled);
			}

			return new Interval(
					Signal.positiveSignalsFromPorts(edc.enablingPorts),
					Interval.SignalCombination.OR,
					Signal.positiveSignalsFromPorts(edc.disablingPorts),
					Interval.SignalCombination.OR,
					edc.initiallyEnabled);
		}

		// the expression is directly a signal, that may be a port of the
		// controller , or a port connected to the controller
		if(clause instanceof SignalPredicate) {
			SignalPredicate sp = (SignalPredicate)clause;
			PortInstance port = sp.getSignal();
			boolean active = sp.isActive();

			// is the port a controller port?
			if(port.getComponent() == controlComponent)
				return new Interval(port, active, port, !active, !active);

			// else, is the port connected to a controller port?
			PortInstance controllerPort = isConnectedWith(port, controlComponent);
			if(controllerPort != null) {
				return new Interval(
						controllerPort, active,
						controllerPort, !active, !active);
			}
		}


		// else: try to find the condition (and its negation) in a when-condition
		// (provided that the expression does not contain temporal operators)
		Expr approximateSolution = null;
		if(! clause.isTemporal()) {
			SignalsEmittedWhen sew = findSignalsEmittedWhen(clause);
			log.debug(sew);
			// if there are begin and end signals...
			if(sew.whenTrue.relations.size() >= 1 || sew.whenFalse.relations.size() >= 1) {
				Set<Signal>
					tEquiv = new HashSet<Signal>(), tPrec = new HashSet<Signal>(),
					tSucc = new HashSet<Signal>(), fEquiv = new HashSet<Signal>(),
					fPrec = new HashSet<Signal>(), fSucc = new HashSet<Signal>();

				sew.whenTrue.listEquivPrecSucc(tEquiv, tPrec, tSucc);
				sew.whenFalse.listEquivPrecSucc(fEquiv, fPrec, fSucc);

				// simple case when there are _equivalent_ begin and end signals
				if(tEquiv.size() > 0 && fEquiv.size() > 0)
					return new Interval(
							tEquiv, Interval.SignalCombination.OR,
							fEquiv, Interval.SignalCombination.OR,
							false);

				// more general case --> build a proto-interval
				Set<Signal> outerBegin, outerEnd, innerBegin, innerEnd;

				if(tEquiv.size() > 0) {
					outerBegin = tEquiv;
					innerBegin = tEquiv;
				} else {
					outerBegin = tSucc;  // outerBegin < interval bound
					innerBegin = tPrec;  // interval bound < innerBegin
				}

				if(fEquiv.size() > 0) {
					outerEnd = fEquiv;
					innerEnd = fEquiv;
				} else {
					outerEnd = fPrec;  // interval bound < outerEnd
					innerEnd = fSucc;  // innerEnd < interval bound
					// note that the relations are reverse with respect to
					// the "beginning" situation (see above)
				}

				Interval inner = null, outer = null;
				if(innerBegin.size() > 0 && innerEnd.size() > 0)
					inner = new Interval(
							innerBegin, Interval.SignalCombination.FIRST,
							innerEnd, Interval.SignalCombination.LAST,
							false);

				if(outerBegin.size() > 0 && outerEnd.size() > 0)
					outer = new Interval(
							outerBegin, Interval.SignalCombination.LAST,
							outerEnd, Interval.SignalCombination.FIRST,
							false);

				if(inner != null || outer != null) {
					// in this case, we have found an approximate solution. We
					// store it for later use, but first try to fallback to
					// the decomposition of the clause into a binary or unary
					// expression. If one of these decompositions succeeds
					// better (i.e. yields an approximation-free result, we use
					// it. Otherwise, we use this approximateSolution
					approximateSolution = new ProtoInterval(inner, outer);
					log.debug("--> proto-interval: " + approximateSolution);
				} // else, if no inner and no outer, fallback to the sequel
			}
		}

		// the expression is of the form "A BinOp B"
		if(clause instanceof BinaryExpr) {
			BinaryExpr be = (BinaryExpr)clause;
			Expr exprA = patternForClause(be.getA(), false);
			Expr exprB = patternForClause(be.getB(), false);

			if(exprA != null && exprB != null)
				///if(approximateSolution == null || (approximateSolution != null && !exprA.isBundled() && !exprB.isBundled()))
					return be.create(exprA, exprB);
		}

		// the expression is of the form "UnaryOp A"
		if(clause instanceof UnaryExpr) {
			UnaryExpr ue = (UnaryExpr)clause;
			Expr expr = patternForClause(ue.getExpr(), false);

			if(expr != null)
				if(approximateSolution == null || !expr.isBundled())
					return ue.create(expr);
		}

		// last resort: our approximate solution
		if(approximateSolution != null)
			return approximateSolution;


		if(abortIfNoSolution)
			log.err("Could not create a pattern for clause: " + clause);
		return null;
	}

	/**
	 * Tests whether a given port is connected with a given component.
	 *
	 * If so, it returns the given component's port that is connected with
	 * the given port. Otherwise, it returns {@code null}.
	 *
	 * Both incoming and outgoing connections are handled.
	 *
	 * @param port a given port
	 * @param component a given component
	 * @return not {@code null} iff the given port is connected with the
	 * given component
	 */
	private static PortInstance isConnectedWith(PortInstance port, Component component) {
		Port pt = port.getPort();
		Component comp = port.getComponent();
		// 1) outgoing connections
		List<Connection> conns = pt.getConnectionFrom();
		for(Connection conn : conns) {
			if(conn.getToComponent() == component && conn.getFromComponent() == comp)
				return PortInstance.make(conn.getToPort(), conn.getToComponent());
		}

		// 2) incoming connections
		conns = pt.getConnectionTo();
		for(Connection conn : conns) {
			if(conn.getFromComponent() == component && conn.getToComponent() == comp)
				return PortInstance.make(conn.getFromPort(), conn.getFromComponent());
		}

		// else, not connected
		return null;
	}


	/**
	 * Generates the checking code for the application. Each "Always" and
	 * "Never" block is translated into checking code.
	 *
	 * This is the class' main entry point.
	 *
	 * @param codeGen a language-specific checking code generator (e.g. Esterel
	 * or Lustre code generator).
	 */
	public void generateCheckingCode(CheckingCodeGenerator codeGen) {
		codeGen.start();
    	for(SafetyClause clause : app.getSafetyClause()) {
    		log.debug("Generate checking code for " +
    				((clause instanceof NeverClause) ? "Never" : "Always") + " clause: " +
    				EmfAdlvHelper.getStringForExpression(clause.getExpression()));

    		Expr clauseExpr = expressionToExpr(clause.getExpression(), null);

    		if(clause instanceof NeverClause)
    			generateCheckingCodeForNeverClause(codeGen, clauseExpr);
    		if(clause instanceof AlwaysClause)
    			generateCheckingCodeForNeverClause(codeGen, new Not(clauseExpr));
    	}

    	log.debug("Finished processing safety clauses.");
    	try {
    		codeGen.stop();
    	} catch(GenerationException e) {
    		// an error occurred during generation
    		log.err(e);
    	}
	}

	public Component getControlComponent() {
		return controlComponent;
	}

	/**
	 * Constructs the list of input and output ports of the controller
	 * @param in an empty list, to be filled with the names of input ports
	 * @param out an empty list, to be filled with the names of output ports
	 */
	public void getControlComponentIO(List<String> in, List<String> out) {
		for(Port p : controlComponent.getType().getPorts()) {
			if(p instanceof ConsumerPort)
				in.add(p.getName());
			if(p instanceof ProducerPort)
				out.add(p.getName());
		}
	}
}


enum BinaryOperatorClassifier {
	COMPARISON, LOGICAL, CONNECTED;

	public static BinaryOperatorClassifier classify(BinaryOperator bo) {
		switch(bo) {
		case EQUAL:
		case NOT_EQUAL:
		case LESS:
		case LESS_EQUAL:
		case GREATER:
		case GREATER_EQUAL:
			return COMPARISON;

		case OR:
		case AND:
		case SINCE:
		case BACK_TO:
		case IMPLIES:
			return LOGICAL;

		case CONNECTED_TO_FROM:
			return CONNECTED;

		default:
			return null;
		}
	}
}