/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.checker;


import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import fr.supelec.adlv.checker.generators.DummyStringGenerator;
import fr.supelec.adlv.checker.generators.EsterelGenerator;
import fr.supelec.adlv.checker.generators.LustreGenerator;
import fr.supelec.adlv.parser.EmfAdlvFeeder;


class Main {
	public static void main(String[] args) throws IOException {
		if(args.length != 6) {
			System.err.println(
					"Syntax: java checker.Main <ADLV file input file> " +
					"<Esterel controller file> <Esterel output file> " +
					"<Lustre controller file> " +
					"<Lustre controller node> <Lustre output file>");
			System.exit(1);
		}

		String inAdlv = args[0];
		String ctrlEsterel = args[1];
		String outEsterel = args[2];
		String ctrlLustre = args[3];
		String ctrlNodeLustre = args[4];
		String outLustre = args[5];

		// create a simple log facility
		LogFacility log = new LogFacility();

		log.info("Input file: " + inAdlv);
		log.info("Esterel Output file: " + outEsterel);
		log.info("Lustre Controller file: " + ctrlLustre);
		log.info("Lustre Controller node: " + ctrlNodeLustre);
		log.info("Lustre Output file: " + outLustre);

		// open ADLV file
		BufferedInputStream input = null;
		try {
			input = new BufferedInputStream( new FileInputStream( inAdlv ));
		}
		catch( FileNotFoundException ex ) {
			log.err( "ADLV file not found: " + ex.toString() );
			return;
		}

		// parse ADLV file and build model
		EmfAdlvFeeder feeder = new EmfAdlvFeeder();
		if( ! feeder.parse( input )) {
			return;
		}

		log.info("Parsing ADLV: done. ");
		String errorStr = feeder.getErrors();
		if("".equals(errorStr)) {
			log.info("No parse error.");
		} else {
			log.err("Parse errors:\n" + errorStr);
		}

		// create the checker generator
		Checker checker = new Checker(feeder.getHelper().getApplication(), log);


		// generate Esterel checking code
		EsterelGenerator eg = new EsterelGenerator(
				inAdlv,
				ctrlEsterel,
				new File(outEsterel),
				checker.getControlComponent());
    	checker.generateCheckingCode(eg);

    	// generate Lustre checking code
    	List<String> inPorts = new ArrayList<String>();
    	List<String> outPorts = new ArrayList<String>();
    	checker.getControlComponentIO(inPorts, outPorts);
    	LustreGenerator lg =
    		new LustreGenerator(inAdlv, inPorts, outPorts, ctrlLustre,
    				ctrlNodeLustre, new File(outLustre), false);
    	checker.generateCheckingCode(lg);

    	// generate dummy strings (Unicode sequences) to standard output
    	checker.generateCheckingCode(new DummyStringGenerator());
	}
}
