/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.checker.generators;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import fr.supelec.adlv.Component;
import fr.supelec.adlv.ConsumerPort;
import fr.supelec.adlv.ProducerPort;
import fr.supelec.adlv.checker.DefaultAbstractCheckingCodeGenerator;
import fr.supelec.adlv.checker.Signal;
import fr.supelec.adlv.checker.Util;
import fr.supelec.adlv.checker.Util.Pair;
import fr.supelec.adlv.checker.logic.Always;
import fr.supelec.adlv.checker.logic.And;
import fr.supelec.adlv.checker.logic.BinaryExpr;
import fr.supelec.adlv.checker.logic.Constant;
import fr.supelec.adlv.checker.logic.Expr;
import fr.supelec.adlv.checker.logic.Interval;
import fr.supelec.adlv.checker.logic.Not;
import fr.supelec.adlv.checker.logic.Once;
import fr.supelec.adlv.checker.logic.Or;
import fr.supelec.adlv.checker.logic.Previous;
import fr.supelec.adlv.checker.logic.Since;


public class EsterelGenerator extends DefaultAbstractCheckingCodeGenerator {
	private final PrintStream ps;
	private final String adlvFileName;
	private final String outputFileName;
	private final Component controlComponent;
	private EsterelModule latestModule = null;
	private final List<SignalRelation> relations = new ArrayList<SignalRelation>();

	private EsterelModule myModule;
	private Set<String> myOutputs = new HashSet<String>();
	private Set<String> mySignals = new HashSet<String>();
	private Map<Expr, EsterelModule> modules = new HashMap<Expr, EsterelModule>();

	public EsterelGenerator(String adlvFileName, String ctrlFileName, File outFile, Component controlComponent) throws IOException {
		ps = new PrintStream(new FileOutputStream(outFile));
		this.controlComponent = controlComponent;
		this.adlvFileName = bareFileName(adlvFileName);
		this.outputFileName = bareFileName(outFile.toString());

		// parsing relations
		//final String moduleStart = "module " + controlComponent.getName() + ":";
		//final String moduleEnd = "end module";
		final Pattern pRelation = Pattern.compile("relation\\s+([\\w\\_]+)\\s*#([\\w\\_]+)\\s*;");
		//boolean inside = false;
		BufferedReader br = new BufferedReader(new FileReader(ctrlFileName));
		while(true) {
			String line = br.readLine();
			if(line == null) break;
			//System.out.println(line);
			//line = line.trim();
			Matcher m = pRelation.matcher(line);
			while(m.find()) {
				if(m.groupCount() != 2) continue;
				relations.add(new SignalRelation(m.group(1), m.group(2), "#"));
				//System.out.println("Found relation: " + relations.get(relations.size()-1));
			}
			//if(line.compareToIgnoreCase(moduleEnd) == 0) break;
			//if(line.compareToIgnoreCase(moduleStart) == 0) inside = true;
		}
	}


	/**
	 * Initiates the creation of an Esterel module based on the parallelization
	 * of two modules.
	 *
	 * @param s
	 * @param moduleA
	 * @param moduleB
	 * @return
	 */
	private EsterelModule initiateEsterelModuleBasedOnTwoModules(EsterelModule moduleA, EsterelModule moduleB) {
		EsterelModule module = new EsterelModule();
		module.inputs.addAll(moduleA.inputs);
		module.inputs.addAll(moduleB.inputs);
		module.relations.addAll(moduleA.relations);
		module.relations.addAll(moduleB.relations);

		module.printHeader(ps);
		ps.println("\tsignal " +
				moduleA.theSignal + ", " + moduleB.theSignal + " in");
		if(moduleA.isReal()) { 		// take optimizations into account
			ps.println("\t\trun " + moduleA.name);
			ps.println("\t\t||");
		}

		if(moduleB.isReal()) { 		// take optimizations into account
			ps.println("\t\trun " + moduleB.name);
			ps.println("\t\t||");
		}

		return module;
	}

	private final Util.StringConverter<Signal> signalToString =
		new Util.StringConverter<Signal>() {
			public String convert(Signal s) {
				String res = s.port.getPort().getName();
				if(s.port.getComponent() != controlComponent)
					throw new RuntimeException("Bug: can't generate Esterel signal for non-controller port " + s.port);
				if(!s.positive) res = "[not " + res + "]";
				return res;
			}
	};

	private final Util.StringConverter<Signal> awaitSignalString =
		new Util.StringConverter<Signal>() {
			public String convert(Signal s) {
				return "[await " + signalToString.convert(s) + "]";
			}
		};

	private Set<String> setOfSignalNames(Set<Signal> set) {
		Set<String> res = new HashSet<String>(set.size());
		for(Signal sig : set) {
			res.add(sig.port.getPort().getName());
		}
		return res;
	}


	private String intervalBound(Set<Signal> signals, Interval.SignalCombination comb) {
		if(signals.size() > 0) {
			switch(comb) {
			case OR:
			case FIRST:
				return "await " + Util.collectionToString(signals, " or ", signalToString);
			case LAST:
				return Util.collectionToString(signals, " || ", awaitSignalString);
			}
		}

		throw new RuntimeException("EsterelGenerator.intervalBound() failed.");
	}

	public void visit(Interval interval) {
		EsterelModule module = new EsterelModule();
		String beginB = intervalBound(interval.getBegin(), interval.getBeginCombination());
		String endB = intervalBound(interval.getEnd(), interval.getEndCombination());

		// add begin signal and end signal
		module.inputs.addAll(setOfSignalNames(interval.getBegin()));
		module.inputs.addAll(setOfSignalNames(interval.getEnd()));

		/*
		 * Note: for an interval of the form [ +sig ; -sig [, sig gets added
		 * only once (module.inputs is a SET).
		 */

		//module.relations.add(new SignalRelation("intv_start", "intv_stop","#"));


		ps.println("% Module for interval: " + interval);
		module.printHeader(ps);

		final String startStr = "\t\t\t[ " + beginB + " ];\n\t\t\temit intv_start;";
		final String endStr = "\t\t\t[ " + endB + " ];\n\t\t\temit intv_stop;";

		ps.println("\tsignal intv_start, intv_stop in");
		ps.println("\t\t[");
		if(interval.getInitialActivationState()) ps.println("\t\temit intv_start");
		ps.println("\t\tloop");
		if(interval.getInitialActivationState()) {
			ps.println(endStr);
			ps.println(startStr);
		} else {
			ps.println(startStr);
			ps.println(endStr);
		}
		ps.println("\t\tend loop");
		ps.println("\t\t] || [");

		if(interval.getInitialActivationState())
			ps.println("\temit " + module.theSignal + ";");

		ps.println("\t\tevery immediate intv_start do");
		ps.println("\t\t\tabort");
		ps.println("\t\t\t\tsustain " + module.theSignal);
		ps.println("\t\t\twhen intv_stop");
		ps.println("\t\tend every");
		ps.println("\t\t]");
		ps.println("\tend signal");

		module.printClose(ps);
		latestModule = module;
	}

	private void visit(BinaryExpr operation, String operator) {
		Expr patternA = operation.getA();
		patternA.accept(this);
		EsterelModule moduleA = latestModule;

		Expr patternB = operation.getB();
		patternB.accept(this);
		EsterelModule moduleB = latestModule;

		ps.println("% Module for \"" + operator + "\" clause: " + operation);
		EsterelModule module = initiateEsterelModuleBasedOnTwoModules(moduleA, moduleB);

		ps.println("\t\t[");
		ps.println("\t\t\tevery immediate [" + moduleA.theSignal + " " +
				operator + " " + moduleB.theSignal + "] do");
		ps.println("\t\t\t\temit " + module.theSignal);
		ps.println("\t\t\tend every");
		ps.println("\t\t]");

		ps.println("\tend signal");
		module.printClose(ps);

		latestModule = module;
	}

	public void visit(And and) {
		visit(and, "and");
	}

	public void visit(Or or) {
		visit(or, "or");
	}

	private Pair<EsterelModule, EsterelModule> initiateEsterelModuleWithSubModule
						(String clauseType, Expr expr, Expr subExpr) {
		Expr subPattern = subExpr;
		subPattern.accept(this);
		EsterelModule subModule = latestModule;

		ps.println("% Module for \"" + clauseType + "\" clause: " + expr);
		EsterelModule module = new EsterelModule();
		module.inputs.addAll(subModule.inputs);
		module.relations = subModule.relations;
		module.printHeader(ps);

		ps.println("\tsignal " + subModule.theSignal + " in");
		ps.println("\t\trun " + subModule.name);
		ps.println("\t\t||");

		return new Pair<EsterelModule, EsterelModule>(module, subModule);
	}


	public void visit(Not not) {
		Pair<EsterelModule, EsterelModule> p =
			initiateEsterelModuleWithSubModule("not", not, not.getExpr());

		EsterelModule module = p.first();
		EsterelModule subModule = p.second();

		ps.println("\t\t[");
		ps.println("\t\t\tevery immediate [ not " + subModule.theSignal + " ] do");
		ps.println("\t\t\t\temit " + module.theSignal);
		ps.println("\t\t\tend every");
		ps.println("\t\t]");

		ps.println("\tend signal");
		module.printClose(ps);

		latestModule = module;
	}


	public void visit(Since since) {
		Expr patternA = since.getA();
		patternA.accept(this);
		EsterelModule moduleA = latestModule;

		Expr patternB = since.getB();
		patternB.accept(this);
		EsterelModule moduleB = latestModule;

		ps.println("% Module for \"since\" clause: " + since);
		EsterelModule module = initiateEsterelModuleBasedOnTwoModules(moduleA, moduleB);

		ps.println("\t\t[");
		ps.println("\t\t\tevery immediate " + moduleB.theSignal + " do");
		ps.println("\t\t\t\tdo");
		ps.println("\t\t\t\t\tsustain " + module.theSignal);
		ps.println("\t\t\t\twatching [not " + moduleA.theSignal + "]");
		ps.println("\t\t\tend every");
		ps.println("\t\t]");

		ps.println("\tend signal");
		module.printClose(ps);

		latestModule = module;
	}


	public void visit(Previous previous) {
		Pair<EsterelModule, EsterelModule> p =
			initiateEsterelModuleWithSubModule("previous", previous, previous.getExpr());

		EsterelModule module = p.first();
		EsterelModule subModule = p.second();

		ps.println("\t\t[");
		ps.println("\t\t\tevery " + subModule.theSignal + " do");
		ps.println("\t\t\t\t; emit " + module.theSignal);
		ps.println("\t\t\tend every");
		ps.println("\t\t]");

		ps.println("\tend signal");
		module.printClose(ps);

		latestModule = module;
	}


	@Override
	public void visit(Once once) {
		Pair<EsterelModule, EsterelModule> p =
			initiateEsterelModuleWithSubModule("once", once, once.getExpr());

		EsterelModule module = p.first();
		EsterelModule subModule = p.second();

		ps.println("\t\t[");
		ps.println("\t\t\tawait " + subModule.theSignal + ";");
		ps.println("\t\t\tsustain " + module.theSignal);
		ps.println("\t\t]");

		ps.println("\tend signal");
		module.printClose(ps);

		latestModule = module;
	}


	public void visit(Always always) {
		Pair<EsterelModule, EsterelModule> p =
			initiateEsterelModuleWithSubModule("always", always, always.getExpr());

		EsterelModule module = p.first();
		EsterelModule subModule = p.second();

		ps.println("\t\t[");
		ps.println("\t\t\tdo");
		ps.println("\t\t\t\tsustain " + module.theSignal);
		ps.println("\t\t\twatching [not " + subModule.theSignal + "]");
		ps.println("\t\t]");

		ps.println("\tend signal");
		module.printClose(ps);

		latestModule = module;
	}


	public void visit(Constant constant) {
		EsterelModule module = new EsterelModule();

		ps.println("% Module for constant: " + constant);
		module.printHeader(ps);

		if(constant.getValue()) ps.println("\tsustain " + module.theSignal);
		else ps.println("\t;");

		module.printClose(ps);
		latestModule = module;
	}


	public void generateNeverClause(Expr pattern, String[] comments) {
		pattern.accept(this);
		EsterelModule module = latestModule;

		// inputs and outputs
		for(Object o : controlComponent.getType().getPorts()) {
			if(o instanceof ConsumerPort) {
				myModule.inputs.add(((ConsumerPort)o).getName());
			}

			if(o instanceof ProducerPort ) {
				myOutputs.add(((ProducerPort)o).getName());
			}
		}

		modules.put(pattern, module);

		// signals
		mySignals.addAll(module.inputs);
		mySignals.add(module.theSignal);

		module.comments = comments;
	}


	private String bareFileName(String path) {
		String[] comps = path.split("/");
		return comps[comps.length - 1];
	}

	public void start() {
		ps.println("% Esterel checker code for " + adlvFileName);
		ps.println("% Generated by MoDriVal checker generator.");
		ps.println("% $Id: EsterelGenerator.java 217 2007-11-27 13:15:06Z jacquet_chr $ ");
		ps.println("%");
		ps.println("% To check, run:");
		ps.println("%    (1) esterel -Lblif -B proof " + outputFileName + " " + controlComponent.getType().getName() + ".strl");
		ps.println("%    (2) checkblif -output failure -rel proof.rel.blif proof.blif");
		ps.println();

		myModule = new EsterelModule("Main_Check_Never");
		myModule.theSignal = "failure";

	}


	public void stop() {
		mySignals.removeAll(myModule.inputs);
		myOutputs.removeAll(mySignals);


		ps.println("% Main module checking the never/always clauses");

		myModule.printHeader(ps);
		if(myOutputs.size() > 0)
			ps.println("\toutput " + Util.collectionToString(myOutputs) + ";");

		for(SignalRelation rel : relations) {
			if(myModule.inputs.contains(rel.a) && myModule.inputs.contains(rel.b))
				ps.println("\trelation " + rel + ";");
		}

		ps.println();

		ps.println("\tsignal " + Util.collectionToString(mySignals) + " in");
		ps.println("\t\trun " + controlComponent.getName());

		for(Map.Entry<Expr, EsterelModule> e : modules.entrySet()) {
			//Expr expr = e.getKey();
			EsterelModule module = e.getValue();

			if(module.comments != null) {
				for(String l : module.comments)
					ps.println("\t\t% " + l);
			}

			ps.println("\t\t||");
			ps.println("\t\trun " + module.name);
			ps.println("\t\t||");
			ps.println("\t\t[");
			ps.println("\t\t\tevery immediate " + module.theSignal + " do");
			ps.println("\t\t\t\temit failure");
			ps.println("\t\t\tend every");
			ps.println("\t\t]");
		}


		ps.println("\tend signal");
		myModule.printClose(ps);

		latestModule = myModule;



		ps.println();
		ps.println("% End of Esterel checker code.");
		ps.close();
	}


	/**
	 * This class models an Esterel module.
	 *
	 */
	private static class EsterelModule {
		public String[] comments = null;
		public String name;
		public String theSignal;
		public final Set<String> inputs = new HashSet<String>();
		public List<SignalRelation> relations = new ArrayList<SignalRelation>();
		private static int esterel_module_id = 1;

		public EsterelModule(String namePrefix) {
			name = namePrefix + "_" + (esterel_module_id++);
			theSignal = "In_" + name;
		}

		public EsterelModule() {
			this("Check");
		}

		public void printDecl(PrintStream s) {
			s.println("module " + name + ":");
		}

		public void printOutputSignals(PrintStream s) {
			s.println("\toutput " + theSignal + ";");
		}

		public void printInputsignals(PrintStream s) {
			if(inputs.size()>0) {
				s.print("\tinput");
				boolean notFirst = false;
				for(String is : inputs) {
					if(notFirst) s.print(",");
					s.print(" " + is);
					notFirst = true;
				}
				s.println(";");
			}
		}

		public void printHeader(PrintStream s) {
			printDecl(s);
			printInputsignals(s);
			printOutputSignals(s);
			s.println();
		}

		public void printClose(PrintStream s) {
			s.println("end module");
			s.println();
		}

		public boolean isReal() {
			return !"NO_MODULE".equals(name);
		}
	}

	public String getName() {
		return "Esterel Generator";
	}
}

/**
 * This class represents a relation between signals.
 *
 */
class SignalRelation {
	public String a, b;
	public String op;

	public SignalRelation(String a, String b, String op) {
		this.a = a;
		this.b = b;
		this.op = op;
	}

	@Override
	public String toString() {
		return a + op + b;
	}
}

