/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.generator;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import fr.supelec.adlv.Application;
import fr.supelec.adlv.Component;
import fr.supelec.adlv.ComponentConfigurationStep;
import fr.supelec.adlv.ComponentType;
import fr.supelec.adlv.ConfigurationStep;
import fr.supelec.adlv.Connection;
import fr.supelec.adlv.ConnectionConfigurationStep;
import fr.supelec.adlv.ConsumerPort;
import fr.supelec.adlv.EventType;
import fr.supelec.adlv.Port;
import fr.supelec.adlv.ProducerPort;
import fr.supelec.adlv.StreamType;
import fr.supelec.adlv.TokenType;
import fr.supelec.adlv.parser.EmfAdlvHelper;

public class CcdGenerator extends AdlvGenerator {

	CcdGenerator( EmfAdlvHelper helper ) {
		helper_ = helper;
	}
	
	void generate( String resultsPath ) throws IOException {
		for( Component c : helper_.getApplication().getComponents() ) {
			generateCcdFileForComponent( c.getType(), resultsPath );
		}
		
		generateCcdFileForDynamicRouting( resultsPath );
		generateCcdFileForEnvironment( resultsPath );
		
		String path = resultsPath + "/" + ENVIRONMENT_NAME;
		
		generateCdpFile( path );
	}
	
	private void generateCcdFileForComponent( ComponentType component, String resultsPath ) throws IOException {
		String cname = component.getName();
		String componentDirectory = resultsPath + "/" + cname;
		( new File( componentDirectory ) ).mkdir();

		openFile( componentDirectory, cname, "ccd" );
		
		HashSet< Port > ports = new HashSet< Port >();
		for( Port p : component.getPorts() ) {
			ports.add( p );
		}
			
		generateCcdFile( cname, ports, false );

		closeFile();
	}
	
	private void generateCcdFileForDynamicRouting( String resultsPath ) throws IOException {
		Application application = helper_.getApplication();
		String applicationDirectory = resultsPath + "/" + DYNAMIC_ROUTING_NAME;
		( new File( applicationDirectory ) ).mkdir();

		openFile( applicationDirectory, DYNAMIC_ROUTING_NAME, "ccd" );
		
		HashSet< Port > usedPorts = new HashSet< Port >();
		for( Port p : application.getPorts() ) {
			// Looking for event consumers
			if( p instanceof ProducerPort ) continue;
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) continue;
			
			// that are used for dynamic configuration
			if((( ConsumerPort ) p ).getSteps().isEmpty() ) continue;
			
			usedPorts.add( p );
			
			for( ConfigurationStep step : (( ConsumerPort ) p ).getSteps() ) {
				if( step instanceof ComponentConfigurationStep ) continue;
				
				usedPorts.add((( ConnectionConfigurationStep ) step ).getConnection().getFromPort() );
				usedPorts.add((( ConnectionConfigurationStep ) step ).getConnection().getToPort() );
			}
		}
		
		generateCcdFile( DYNAMIC_ROUTING_NAME, usedPorts, true );
		
		closeFile();
	}
	
	private void generateCcdFileForEnvironment( String resultsPath ) throws IOException {
		String applicationDirectory = resultsPath + "/" + ENVIRONMENT_NAME;
		( new File( applicationDirectory ) ).mkdir();

		openFile( applicationDirectory, ENVIRONMENT_NAME, "ccd" );
		
		generateCcdFileHeader( ENVIRONMENT_NAME );
		
		HashSet< Port > ports = new HashSet< Port >();
		ports.addAll( helper_.getApplicationExternalConsumerPorts() );
		ports.addAll( helper_.getApplicationExternalProducerPorts() );
		generateCcdFilePorts( ports, false );
		
		// We also provide the internal consumer ports (those used for dynamic
		// configuration) so that they can be examined
		ports.clear();
		ports.addAll( helper_.getApplicationInternalConsumerPorts() );
		generateCcdFilePorts( ports, true );
		
		generateCcdFileFooter();
		
		closeFile();
	}
	
	private void generateCcdFileHeader( String name ) throws IOException {
		out_.write( "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n" );

		writeXmlCopyrightComment();

		out_.write( "\n" );
		out_.write( "<ComponentInterfaceDescription>\n" );
		out_.write( "    <label>"                       + name + "</label>\n" );
		out_.write( "    <UUID>"                        + name + "</UUID>\n" );
		//out_.write( "    <specificType>" + mname + "::" + name + "</specificType>\n" );
		out_.write( "    <specificType>"                + name + "</specificType>\n" );
		out_.write( "    <supportedType>"                       + "</supportedType>\n" );
		out_.write( "    <idlFile>../"  + name  + "/"  + name + ".idl3</idlFile>\n" );
		out_.write( "    <configProperty>"                      + "</configProperty>\n" );
	}
	
	private void generateCcdFileFooter( ) throws IOException {
		out_.write( "</ComponentInterfaceDescription>\n" );
		out_.write( "\n" );
	}
	
	private void generateCcdFilePorts( HashSet< Port > ports, boolean forApplication ) throws IOException {
		for( Port p : ports ) {
			TokenType t = p.getTokenType();
			out_.write( "    <port>\n" );
			out_.write( "        <name>" + p.getName() + "</name>\n" );
			if( t instanceof StreamType ) {
				StreamType s = (StreamType) t;
				out_.write( "        <specificType>" + getStreamType( s.getType().getName() ) + "</specificType>\n" );
			}
			else {
				out_.write( "        <specificType>" + t.getModule().getName() + "::" + t.getName() + "</specificType>\n" );
			}
			out_.write( "        <supportedType>" + "</supportedType>\n" );
			out_.write( "        <exclusiveProvider>" + "false" + "</exclusiveProvider>\n" );
			out_.write( "        <exclusiveUser>" + "false" + "</exclusiveUser>\n" );
			out_.write( "        <optional>" + "true" + "</optional>\n" );
			boolean inverseRole = forApplication;
			if( p.getApplication() != null ) inverseRole = ! inverseRole;
			if(( inverseRole && ( p instanceof ProducerPort )) || (( ! inverseRole ) && ( p instanceof ConsumerPort ))) {
				// In CCM, EventConsumer and Facet ports seem to be considered providers, the other ports users
				// (OMG doc formal/06-04-02, section 10.3.1).
				// In current version of Inflexion, EventPublisher are considered provider
				//out_.write( "        <provider>" + "true" + "</provider>\n" );
				if( t instanceof StreamType ) {
					out_.write( "        <provider>" + "true" + "</provider>\n" );
					out_.write( "        <kind>" + "Facet" + "</kind>\n" );
				}
				else {
					out_.write( "        <provider>" + "false" + "</provider>\n" );
					out_.write( "        <kind>" + "EventConsumer" + "</kind>\n" );
				}
			}
			else {
				//out_.write( "        <provider>" + "false" + "</provider>\n" );
				if( t instanceof StreamType ) {
					out_.write( "        <provider>" + "false" + "</provider>\n" );
					out_.write( "        <kind>" + "SimplexReceptacle" + "</kind>\n" );
				}
				else {
					out_.write( "        <provider>" + "true" + "</provider>\n" );
					out_.write( "        <kind>" + "EventPublisher" + "</kind>\n" );
				}
			}
			out_.write( "    </port>\n" );
		}
	}
	
	private void generateCcdFile( String name, HashSet< Port > ports, boolean forApplication ) throws IOException {
		generateCcdFileHeader( name );
		generateCcdFilePorts( ports, forApplication );
		generateCcdFileFooter();
	}
	
	private void generateCdpFile( String path ) throws IOException {
		String name = SIMULATION_NAME;
		openFile( path, name, "cdp" );
		
		out_.write( "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n" );

		writeXmlCopyrightComment();

		out_.write( "<Deployment\n" );
		out_.write( "  xmlns=\"http://www.omg.org/Deployment\"\n" );
		out_.write( "  xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">\n" );
		out_.write( "    <DeploymentPlan>\n" );
		out_.write( "        <label>" + name + "</label>\n" );
		out_.write( "        <UUID>" + name + "</UUID>\n" );
		out_.write( "        <realizes></realizes>\n" );
		
		HashSet< ComponentType > componentTypes = new HashSet< ComponentType >();
		for( Component c : helper_.getApplication().getComponents() ) {
			componentTypes.add( c.getType() );
		}
		for( ComponentType c : componentTypes ) {
			generateCdpForComponentType( c );
		}
		generateCdpImplementation( DYNAMIC_ROUTING_NAME, DYNAMIC_ROUTING_NAME );
		generateCdpImplementation( ENVIRONMENT_NAME, ENVIRONMENT_NAME );
		generateCdpImplementation( "Supelec", "BooleanStreamDuplicator" );
		generateCdpImplementation( "Supelec", "DoubleStreamDuplicator" );
		generateCdpImplementation( "Supelec", "LongStreamDuplicator" );
		
		for( Component c : helper_.getApplication().getComponents() ) {
			generateCdpInstance( c );
		}
		generateCdpInstance( DYNAMIC_ROUTING_NAME, DYNAMIC_ROUTING_NAME );
		generateCdpInstance( ENVIRONMENT_NAME, ENVIRONMENT_NAME );
		
		generateCdpConnections();

		out_.write( "    </DeploymentPlan>\n" );
		out_.write( "</Deployment>\n" );
		out_.write( "\n" );
		
		closeFile();
	}
	
	private void generateCdpForComponentType( ComponentType component ) throws IOException {
		String name = component.getName();
		generateCdpImplementation( name, name );
	}
		
	private void generateCdpImplementation( String dir, String name ) throws IOException {
		out_.write( "        <implementation id=\"" + name + "_implementation\">\n" );
		out_.write( "            <name>" + name + "</name>\n" );
		out_.write( "            <source>../" + dir + "/" + name + ".ccd</source>\n" );
		out_.write( "            <artifact ref=\"" + name + "_artifact\"/>\n" );
		out_.write( "        </implementation>\n" );
		
		out_.write( "        <artifact id=\"" + name + "_artifact\">\n" );

		if( STATIC_DEPLOYMENT ) {
			if( STATIC_LIBRARIES ) {
				out_.write( "            <name>../" + dir + "/lib" + name + "Impl_gcc_linux_tao.a</name>\n" );
				out_.write( "            <location></location>\n" );
			}
			else {
				out_.write( "            <name>lib" + name + "Impl_gcc_linux_tao.so</name>\n" );
				out_.write( "            <location></location>\n" );
			}
		}
		else if ( CORBA_DEPLOYMENT ) {
			if( STATIC_LIBRARIES ) {
				// TODO
			}
			else {
				out_.write( "            <name>../" + dir + "/lib" + name + "Impl_gcc_linux_tao.a</name>\n" );
				out_.write( "            <name>lib" + name + "Impl_gcc_linux_tao.so</name>\n" );
				out_.write( "            <location>tftp://localhost/lib" + name + "Impl_gcc_linux_tao.so</location>\n" );
			}
		}

		out_.write( "            <node></node>\n" );
		out_.write( "            <source>" + name + "</source>\n" );
		out_.write( "            <execParameter>\n" );
		out_.write( "                <name>home factory</name>\n" );
		out_.write( "                <value>create_" + name + "_home</value>\n" );
		out_.write( "            </execParameter>\n" );
		out_.write( "        </artifact>\n" );
	}
	
	private void generateCdpInstance( String name, String iname ) throws IOException {
		out_.write( "        <instance id=\"" + name + "_instance\">\n" );
		out_.write( "            <name>" + name + "</name>\n" );
		out_.write( "            <node>" + "localhost" + "</node>\n" );
		out_.write( "            <implementation ref=\"" + iname + "_implementation\"/>\n" );
		out_.write( "        </instance>\n" );
	}
	private void generateCdpInstance( Component component ) throws IOException {
		generateCdpInstance( component.getName(), component.getType().getName() );
	}
	
	private void generateCdpConnections() throws IOException {
		HashSet< Connection > connections = new HashSet< Connection >();
		
		// Static connections
		for( ConfigurationStep step : helper_.getApplication().getStaticConfiguration() ) {
			if( step instanceof  ComponentConfigurationStep ) continue;
			
			connections.add((( ConnectionConfigurationStep ) step ).getConnection());
		}
		
		for( Connection conn : connections ) {
			Component c = conn.getFromComponent();
			Port p = conn.getFromPort();
			newConsumerFor( c, p );
		}
		
		ArrayList< Connection > connectionsToDuplicate = new ArrayList< Connection >();
		for( Connection conn : connections ) {
			Component c = conn.getFromComponent();
			Port p = conn.getFromPort();
			if( getNbConsumersFor( c, p ) > 1 ) {
				// Events are correctly duplicated
				if( p.getTokenType() instanceof StreamType ) {
					connectionsToDuplicate.add( conn );
				}
			}
		}
		connections.removeAll( connectionsToDuplicate );
		
		int numConnection = 0;
		for( Connection conn : connections ) {
			String name = "Connexion_" + ++numConnection;
			out_.write( "        <connection id=\"" + name + "_connection\">\n" );
			out_.write( "            <name>" + name + "</name>\n" );

			String ifname;
			Component c = conn.getFromComponent();
			if( c != null ) ifname = c.getName();
			else if( helper_.isApplicationInternalPort( conn.getFromPort().getName() )) {
				ifname = DYNAMIC_ROUTING_NAME;
			}
			else ifname = ENVIRONMENT_NAME;
			String pfname = conn.getFromPort().getName();
			boolean isStream = conn.getFromPort().getTokenType() instanceof StreamType;
			generateCdpInternalPoint( ifname, pfname, false, isStream );
			
			String itname;
			c = conn.getToComponent();
			if( c != null ) itname = c.getName();
			else if( helper_.isApplicationInternalPort( conn.getToPort().getName() )) {
				itname = DYNAMIC_ROUTING_NAME;
			}
			else itname = ENVIRONMENT_NAME;
			String ptname = conn.getToPort().getName();
			generateCdpInternalPoint( itname, ptname, true, isStream );
			
			out_.write( "        </connection>\n" );
			
			// If it is a connection to the dynamic routing component, we also
			// made a connection to the environment
			if( itname == DYNAMIC_ROUTING_NAME ) {
				itname = ENVIRONMENT_NAME;
				name = "Connexion_" + ++numConnection;
				out_.write( "        <connection id=\"" + name + "_connection\">\n" );
				out_.write( "            <name>" + name + "</name>\n" );

				isStream = conn.getFromPort().getTokenType() instanceof StreamType;
				generateCdpInternalPoint( ifname, pfname, false, isStream );
				generateCdpInternalPoint( itname, ptname, true, isStream );
				
				out_.write( "        </connection>\n" );
			}
		}
		
		int numDuplicator = 0;
		while( ! connectionsToDuplicate.isEmpty() ) {
			Connection conn = connectionsToDuplicate.get( 0 );
			connectionsToDuplicate.remove( 0 );
			Component cFrom = conn.getFromComponent();
			Port pFrom = conn.getFromPort();
			TokenType t = pFrom.getTokenType();
			if( t instanceof EventType ) {
				// TODO
				continue;
			}
			String dname = "Duplicator_" + ++numDuplicator;
			StreamType s = (StreamType) t;
			out_.write( "        <instance id=\"" + dname + "_instance\">\n" );
			out_.write( "            <name>" + dname + "</name>\n" );
			out_.write( "            <node>" + "localhost" + "</node>\n" );
			out_.write( "            <implementation ref=\"" + getDuplicatorName( s.getType().getName() ) + "_implementation\"/>\n" );
			out_.write( "        </instance>\n" );
			
			String name = "Connexion_" + ++numConnection;
			out_.write( "        <connection id=\"" + name + "_connection\">\n" );
			out_.write( "            <name>" + name + "</name>\n" );
			String iname;
			Component c = conn.getFromComponent();
			if( c != null ) iname = c.getName();
			else            iname = ENVIRONMENT_NAME;
			String pname = conn.getFromPort().getName();
			generateCdpInternalPoint( iname, pname, false, true );
			generateCdpInternalPoint( dname, "input", true, true );
			out_.write( "        </connection>\n" );

			name = "Connexion_" + ++numConnection;
			out_.write( "        <connection id=\"" + name + "_connection\">\n" );
			out_.write( "            <name>" + name + "</name>\n" );
			c = conn.getToComponent();
			if( c != null ) iname = c.getName();
			else            iname = ENVIRONMENT_NAME;
			pname = conn.getToPort().getName();
			generateCdpInternalPoint( dname, "output_1", false, true );
			generateCdpInternalPoint( iname, pname, true, true );
			out_.write( "        </connection>\n" );

			int nbConsumersLeft = getNbConsumersFor( cFrom, pFrom ) - 1;
			int i = 0;
			while( i < connectionsToDuplicate.size() ) {
				Connection conn2 = connectionsToDuplicate.get( i );
				Component cFrom2 = conn2.getFromComponent();
				Port pFrom2 = conn2.getFromPort();
				
				if(( cFrom2 == cFrom ) && ( pFrom2 == pFrom )) {
					connectionsToDuplicate.remove( i );
					if( nbConsumersLeft > 1 ) {
						String dname2 = "Duplicator_" + ++numDuplicator;
						out_.write( "        <instance id=\"" + dname2 + "_instance\">\n" );
						out_.write( "            <name>" + dname2 + "</name>\n" );
						out_.write( "            <node>" + "localhost" + "</node>\n" );
						out_.write( "            <implementation ref=\"" + getDuplicatorName( s.getType().getName() ) + "_implementation\"/>\n" );
						out_.write( "        </instance>\n" );
						
						name = "Connexion_" + ++numConnection;
						out_.write( "        <connection id=\"" + name + "_connection\">\n" );
						out_.write( "            <name>" + name + "</name>\n" );
						generateCdpInternalPoint( dname, "output_2", false, true );
						generateCdpInternalPoint( dname2, "input", true, true );
						out_.write( "        </connection>\n" );
						dname = dname2;

						name = "Connexion_" + ++numConnection;
						out_.write( "        <connection id=\"" + name + "_connection\">\n" );
						out_.write( "            <name>" + name + "</name>\n" );
						c = conn2.getToComponent();
						if( c != null ) iname = c.getName();
						else            iname = ENVIRONMENT_NAME;
						pname = conn2.getToPort().getName();
						generateCdpInternalPoint( dname, "output_1", false, true );
						generateCdpInternalPoint( iname, pname, true, true );
						out_.write( "        </connection>\n" );
					}
					else {
						name = "Connexion_" + ++numConnection;
						out_.write( "        <connection id=\"" + name + "_connection\">\n" );
						out_.write( "            <name>" + name + "</name>\n" );
						c = conn2.getToComponent();
						if( c != null ) iname = c.getName();
						else            iname = ENVIRONMENT_NAME;
						pname = conn2.getToPort().getName();
						generateCdpInternalPoint( dname, "output_2", false, true );
						generateCdpInternalPoint( iname, pname, true, true );
						out_.write( "        </connection>\n" );
					}
					
					--nbConsumersLeft;
				}
				else {
					++i;
				}
			}
		}
		
		// Dynamic connections through the application component
		// We must group them according to the destination port
		HashMap< Port, HashSet< Connection > > dynamicConnections = new HashMap< Port, HashSet< Connection > >();
		for( Port p : helper_.getApplication().getPorts() ) {
			// Looking for event consumers
			if( p instanceof ProducerPort ) continue;
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) continue;
			
			// that are used for dynamic configuration
			if((( ConsumerPort ) p ).getSteps().isEmpty() ) continue;
			
			for( ConfigurationStep step : (( ConsumerPort ) p ).getSteps() ) {
				if( step instanceof ComponentConfigurationStep ) continue;
				
				Connection conn = (( ConnectionConfigurationStep ) step ).getConnection();
				if( ! dynamicConnections.containsKey( conn.getToPort() )) {
					dynamicConnections.put( conn.getToPort() , new HashSet< Connection >() );
				}
				dynamicConnections.get( conn.getToPort() ).add( conn );
			}
		}
		
		for( Port dest : dynamicConnections.keySet() ) {
			Component to = null;
			for( Connection conn : dynamicConnections.get( dest )) {
				if(( to == null ) && ( conn.getToComponent() != null )) {
					to = conn.getToComponent();
				}
				String name = "Connexion_" + ++numConnection;
				out_.write( "        <connection id=\"" + name + "_connection\">\n" );
				out_.write( "            <name>" + name + "</name>\n" );
				String cname;
				if( conn.getFromComponent() != null ) cname = conn.getFromComponent().getName();
				else cname = ENVIRONMENT_NAME;
				String aname = DYNAMIC_ROUTING_NAME;
				String pname = conn.getFromPort().getName();
				generateCdpInternalPoint( cname, pname, false, true );
				generateCdpInternalPoint( aname, pname, true, true );
				out_.write( "        </connection>\n" );
			}
			String name = "Connexion_" + ++numConnection;
			out_.write( "        <connection id=\"" + name + "_connection\">\n" );
			out_.write( "            <name>" + name + "</name>\n" );
			String cname;
			if( to != null ) cname = to.getName();
			else cname = ENVIRONMENT_NAME;
			String aname = DYNAMIC_ROUTING_NAME;
			String pname = dest.getName();
			generateCdpInternalPoint( aname, pname, false, true );
			generateCdpInternalPoint( cname, pname, true, true );
			out_.write( "        </connection>\n" );
		}
		
	}
	
	private void generateCdpInternalPoint( String iname, String pname, boolean isProvider, boolean isStream ) throws IOException {
		out_.write( "            <internalEndpoint>\n" );
		out_.write( "                <instance ref=\"" + iname + "_instance\"/>\n" );
		out_.write( "                <portName>" + pname + "</portName>\n" );
		// See comment about provider in generateCcdFile()
		if( isProvider ) {
			//out_.write( "                <provider>" + "true" + "</provider>\n" );
			out_.write( "                <provider>" + ( isStream ? "true" : "false" ) + "</provider>\n" );
			out_.write( "                <kind>" + ( isStream ? "Facet" : "EventConsumer" ) + "</kind>\n" );
		}
		else {
			//out_.write( "                <provider>" + "false" + "</provider>\n" );
			out_.write( "                <provider>" + ( isStream ? "false" : "true" ) + "</provider>\n" );
			out_.write( "                <kind>" + ( isStream ? "SimplexReceptacle" : "EventPublisher" ) + "</kind>\n" );
		}
		out_.write( "            </internalEndpoint>\n" );
	}
	
	private int newConsumerFor( Component c, Port p ) {
		int nbConsumers = 0;
		if( c != null ) {
			if( ! fromC_.containsKey( c )) {
				fromC_.put( c, new HashMap< Port, Integer >() );
			}
			if( ! fromC_.get( c ).containsKey( p )) {
				fromC_.get( c ).put( p, new Integer( 0 ));
			}
			nbConsumers = fromC_.get( c ).get( p ) + 1;
			fromC_.get( c ).put( p, nbConsumers );
		}
		else {
			if( ! fromA_.containsKey( p )) {
				fromA_.put( p, new Integer( 0 ));
			}
			nbConsumers = fromA_.get( p ) + 1;
			fromA_.put( p, nbConsumers );
		}
		
		return nbConsumers;
	}
	
	private int getNbConsumersFor( Component c, Port p ) {
		int nbConsumers = 0;
		if( c != null ) {
			if( ! fromC_.containsKey( c )) {
				// TODO : error
			}
			if( ! fromC_.get( c ).containsKey( p )) {
				// TODO : error
			}
			nbConsumers = fromC_.get( c ).get( p );
		}
		else {
			if( ! fromA_.containsKey( p )) {
				// TODO : error
			}
			nbConsumers = fromA_.get( p );
		}
		
		return nbConsumers;
	}
	
	private EmfAdlvHelper helper_;
	HashMap< Component, HashMap< Port, Integer > > fromC_ = new HashMap< Component, HashMap <Port, Integer > >();
                        HashMap< Port, Integer >   fromA_ = new HashMap< Port, Integer >();
}
